##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
# 
##############################################################################
"""

Revision information:
$Id: EventService.py,v 1.1 2003/03/12 10:57:56 smigniot Exp $
"""

from IEventService import IEventService
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo
import Globals
from EventPermissions import *
from Persistence import Persistent
from Interface.Implements import objectImplements
from BTrees import OOBTree
import types
from ISubscribable import ISubscribable
from ISubscriptionAware import ISubscriptionAware
from IEvent import IEvent

def removeAllProxies(ob):
    return getattr(ob, 'aq_base', ob)

class NotFoundError(Exception): pass

class EventService(SimpleItem):
    __implements__ = IEventService

    meta_type = 'Event Service'
    security = ClassSecurityInfo()

    ok = {'meta_type':1, 'id':1, 'icon':1, 'bobobase_modification_time':1 ,
          'title_or_id': 1}
    security.setDefaultAccess(ok)
    security.setPermissionDefault(PUBLISH_EVENTS_PERM, ['Manager'])
    security.setPermissionDefault(SUBSCRIBE_EVENTS_PERM, ['Manager'])
    manage_options = (
        {'label': 'View Subscriptions', 'action': 'manage_viewSubscriptions'},
        )

    def __init__(self):
        self._registry = TypeRegistry()

    security.declareProtected(PUBLISH_EVENTS_PERM, 'publishEvent')
    def publishEvent(self, event, context=None):
        # we need a context for the subscriber.  If one
        # is not passed in, we nominate ourselves as the context.
        if context is None:
            context = self 
        subscriptions_collection = self._registry.getAllForObject(event)

        for subscriptions in subscriptions_collection:
            for subscriber,filter in subscriptions:
                if filter is not None:
                    if not filter(event.__of__(context)):
                        continue
                subscriber = subscriber.__of__(context)
                subscriber.notify(event)

    security.declareProtected(MGMT_SCREEN_PERM, 'listAllSubscriptions')
    def listAllSubscriptions(self):
        l = []
        for interface, stuff in self._registry.all():
            for subscriber, filter in stuff:
                subscriptions = self.listSubscriptions(subscriber, interface)
                for subscription, filter in subscriptions:
                    l.append((subscriber.__of__(self), subscription, filter))
        return l

    def subscribe(self, subscriber, event_type=IEvent, filter=None):
        clean_subscriber=removeAllProxies(subscriber)
        
        if ISubscriptionAware.isImplementedBy(subscriber):
            subscriber.subscribedTo(self, event_type, filter)
        
        ev_type=event_type
        if ev_type is IEvent: ev_type=None # optimization
        
        subscribers = self._registry.getJustForType(ev_type)
        if subscribers is None:
            subscribers = []
        subscribers.append((clean_subscriber, filter))
        self._registry.register(ev_type, subscribers)

    def unsubscribe(self, subscriber, event_type=None, filter=None):
        
        clean_subscriber=removeAllProxies(subscriber)

        do_alert=ISubscriptionAware.isImplementedBy(clean_subscriber)
        
        ev_type=event_type
        if event_type is IEvent:
            ev_type=None # handle optimization
        subscriptions = self._registry.getJustForType(ev_type)
        if not subscriptions:
            raise NotFoundError(subscriber, event_type, filter)
        try: 
            subscriptions.remove((clean_subscriber, filter))
            self._registry.register(ev_type, subscriptions)
        except ValueError:
            raise NotFoundError(subscriber, event_type, filter)
        if do_alert:
            subscriber.unsubscribedFrom(self, event_type, filter)

    def listSubscriptions(self, subscriber, event_type=None):
        
        subscriber=removeAllProxies(subscriber)
        
        result=[]
        ev_type=event_type
        if event_type is IEvent:
            ev_type=None # handle optimization
        subscriptions = self._registry.getJustForType(ev_type)
        if subscriptions:
            for stored_subscriber, stored_filter in subscriptions:
                if stored_subscriber==subscriber:
                    result.append((event_type, stored_filter))
        return result

    security.declareProtected(MGMT_SCREEN_PERM, 'manage_viewSubscriptions')
    manage_viewSubscriptions = PageTemplateFile(
        'www/view_subscriptions',
        globals(),
        __name__='manage_viewSubscriptions'
        )

    # proxy security declarations for base classes
    security.declareProtected(SUBSCRIBE_EVENTS_PERM, 'subscribe')
    security.declareProtected(SUBSCRIBE_EVENTS_PERM, 'unsubscribe')
    security.declareProtected(SUBSCRIBE_EVENTS_PERM, 'listSubscriptions')

class TypeRegistry(Persistent):
    """
    Simple type registry for event service (just the required functionality).
    """
    def __init__(self):
        self._reg = {}

    def update(self):
        if isinstance(self._reg, types.DictionaryType):
            self._p_changed = 1

    def register(self, interface, data):
        self._reg[interface] = data
        self.update()

    def unregister(self, interface):
        del self._reg[interface]
        self.update()

    def all(self):
        return list(self._reg.items())

    def getJustForType(self, interface):
        return self._reg.get(interface, None)

    def getAllForObject(self, ob):
        result = []
        seen = {}
        for i in objectImplements(removeAllProxies(ob)):
            self.getAll(i, result, seen)

        return self.getAll(None, result, seen)

    def getAll(self, interface, result=None, seen=None):
        if result is None:
            result = []
        if seen is None:
            seen = {}

        if seen.has_key(interface):
            return result

        c = self._reg.get(interface, None)
        if c is not None:
            result.append(c)
        seen[interface] = 1
        bases = getattr(interface, '__bases__', ())
        if bases:
            for base in bases:
                self.getAll(base, result, seen)

        return result

def removeAllProxies(ob):
    return getattr(ob, 'aq_base', ob)
    
manage_addEventServiceForm = PageTemplateFile(
        'www/addEventService',
        globals(),
        __name__='manage_addEventServiceForm'
        )

def manage_addEventService(self, id, title, REQUEST=None):
    """ """
    ob = EventService()
    ob.id = id
    ob.title = title
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)

Globals.InitializeClass(EventService)
