# Copyright (c) 2002-2003 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2002 Prfecture du Bas-Rhin, France
# Copyright (c) 2002 CIRB, Belgique
# Authors: Florent Guillaume <mailto:fg@nuxeo.com>
#          Julien Jalon <mailto:jj@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: BaseDocument.py,v 1.18 2003/04/01 21:53:58 fermigier Exp $

"""
  BaseDocument
"""

import string
from urllib import quote

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from Acquisition import aq_base

from OFS.ObjectManager import ObjectManager
from OFS.PropertyManager import PropertyManager
from OFS.FindSupport import FindSupport

from Products.CMFCore.utils import _verifyActionPermissions
from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent
from Products.CMFCore.PortalContent import PortalContent

from Products.VersionManager.VersionManager import VersionManager
from Products.PortalContentFolder.PortalContentFolder import PortalContentFolder
from Products.NuxHierarchies.ReviewedPublication import ReviewedPublication
from Products.NuxExtendedWorkflow.WorkflowAware import WorkflowAware
#from Products.NuxExtendedWorkflow.WorkflowManager import WorkflowManager


factory_type_information = (
    {'id': 'Base Document',
     'description': 'A Base Document is a base type for new document types.',
     'meta_type': 'Base Document',
     'content_icon': 'document_icon.gif',
     'product': 'NuxCPSDocuments',
     'factory': 'addBaseDocument',
     'immediate_view': 'CHANGE_document_edit_form',
     'allow_discussion': 1,
     'filter_content_types': 0,
     'actions': ({'id': 'view',
                  'name': '_action_view_',
                  'action': 'CHANGE_document_view',
                  'permissions': (View,)},
                 {'id': 'edit',
                  'name': '_action_modify_',
                  'action': 'CHANGE_document_edit_form',
                  'permissions': (ModifyPortalContent,)},
                 {'id': 'create',
                  'name': '_action_create_',
                  'action': 'CHANGE_document_create_form',
                  'visible': 0,
                  'permissions': ()},
                 {'id': 'isdocument',
                  'name': 'isdocument',
                  'action': 'isdocument',
                  'visible': 0,
                  'permissions': ()},
                 {'id': 'issearchabledocument',
                  'name': 'issearchabledocument',
                  'action': 'issearchabledocument',
                  'visible': 0,
                  'permissions': ()},
                 ),
     },
    )


class BaseDocument(VersionManager, WorkflowAware, ReviewedPublication,
                   PortalContentFolder, PropertyManager):
    """Base class for CPS documents."""

    meta_type = 'Base Document'

    security = ClassSecurityInfo()

    manage_options = (PropertyManager.manage_options[:1] + # Properties
                      ObjectManager.manage_options[:1] + # Contents
                      PortalContent.manage_options + # Dublin, Edit, View, Own, Sec
                      FindSupport.manage_options # Find
                      )

    _isDiscussable = 1
    isPrincipiaFolderish = 0

    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w', 
         'label': 'Title'},
        {'id': 'description', 'type': 'text', 'mode': 'w', 
         'label': 'Description'},
        {'id': 'related_links', 'type': 'lines', 'mode' :'w', 
         'label': 'Related links'},
        )
    title = ''
    description = ''
    related_links = []

    effective_date = expiration_date = None
    last_modifier = None

    def __init__(self, id, **kw):
        PortalContentFolder.__init__(self, id)
        self.manage_changeProperties(**kw)

    #
    # Dublin Core: explicit treatment of last modifier
    #
    security.declarePrivate('setLastModifier')
    def setLastModifier(self):
        """Change the last modifier's id."""
        userid = getSecurityManager().getUser().getUserName()
        # XXX also store user folder path, like Owner does ?
        self.last_modifier = userid

    security.declarePublic('getLastModifier')
    def getLastModifier(self):
        """Get the the last modifier's id.  """
        return self.last_modifier or self.Creator()

    # Overloaded
    security.declarePrivate('notifyModified')
    def notifyModified(self):
        """
        To be called after the document has been modified.
        This changes the modification date an the modifier name.
        """
        self.setModificationDate()
        self.setLastModifier()

    #
    # Overload absolute_url to take into account virtual URLs
    # when accessing documents through hierarchies.
    #
    security.declarePublic('absolute_url')
    def absolute_url(self, relative=0, munge=1):
        try:
            req = self.REQUEST
        except AttributeError:
            req = {}
        rpp = req.get('VirtualRootPhysicalPath', ('',))
        spp = self.getPhysicalPath()
        ###
        # maybe munge absolute url
        if munge:
            info = req.get('munge_absolute_url', None)
            if info is not None:
                (realpath, virtualpath) = info
                lrp = len(realpath)
                # if begins with realpath, substitute virtualpath
                if spp[:lrp] == realpath:
                    spp = virtualpath + spp[lrp:]
        #
        ###
        i = 0
        for name in rpp[:len(spp)]:
            if spp[i] == name:
                i = i + 1
            else:
                break
        path = map(quote, spp[i:])
        if relative:
            # This is useful for physical path relative to a VirtualRoot
            return string.join(path, '/')
        return string.join([req['SERVER_URL']] + req._script + path, '/')

    #
    # Method to be overloaded by derived products, but generic enough
    # to be usable for simple skin products.
    #
    security.declareProtected(ModifyPortalContent, 'edit')
    def edit(self, **kw):
        """Edit the document."""
        self.manage_changeProperties(**kw)
        self.reindexObject()

    security.declareProtected(View, 'SearchableText')
    def SearchableText(self, tstring=type(''), tseq=(type(()),type([]))):
        """Return a string with all string value properties.

        Used by the catalog for basic full text indexing.
        """
        values = self.propertyValues()
        ok = []
        for val in values:
            t = type(val)
            if t is tstring:
                ok.append(val)
                continue
            if t in tseq:
                for v in val:
                    if type(v) is tstring:
                        ok.append(v)
        return string.join(ok, ' ')

    #
    # Method that calls the Types Tool action
    #
    security.declareProtected(View, 'render')
    def render(self, *args, **kw):
        """Call the render action."""
        ti = self.getTypeInfo()
        if ti is not None:
            actions = ti.getActions()
            for action in actions:
                if action.get('id', None) == 'render':
                    if _verifyActionPermissions(self, action):
                        render = self.restrictedTraverse(action['action'])
                        if getattr(aq_base(render), 'isDocTemp', 0):
                            return render(self, self.REQUEST, *args, **kw)
                        else:
                            return render(*args, **kw)
        raise 'Not Found', ('Cannot find render action for "%s"' %
                            string.join(self.getPhysicalPath(), '/'))

    security.declareProtected(View, 'getRelatedLinks')
    def getRelatedLinks(self, portal_url=''):
        # compatibility
        links_field = self.related_links
        links_field = [line.strip() for line in links_field]
        links_field = filter(None, links_field)

        links = []
        for link in links_field:
            space_position = link.find(' ')
            if space_position > -1:
                linkurl = link[:space_position]
                linktitle = link[space_position+1:].strip()
            else:
                linkurl = link
                linktitle = ''
            linkurl = linkurl.replace('$portal_url$', portal_url)
            links.append((linkurl, linktitle))
        return links


InitializeClass(BaseDocument)


def addBaseDocument(dispatcher, id, REQUEST=None, **kw):
    """Add a Base Document"""
    ob = BaseDocument(id, **kw)
    return BaseDocument_adder(dispatcher, id, ob, REQUEST=REQUEST)

def BaseDocument_adder(dispatcher, id, ob, REQUEST=None, **kw):
    """Add a just constructed object to its container."""
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)


