# Copyright 2002-2003 Nuxeo SARL <http://nuxeo.com>
# Copyright 2002-2003 CEA, France
# Author: Sbastien Migniot <smigniot@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: Newsletter.py,v 1.9 2003/07/14 21:06:47 fermigier Exp $

"""
  Newsletters.
"""

from zLOG import LOG, PROBLEM, ERROR, BLATHER, DEBUG, INFO
from Globals import InitializeClass
from Acquisition import aq_parent
from AccessControl import ClassSecurityInfo, getSecurityManager
from DateTime.DateTime import DateTime

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent, \
    ManagePortal

from BaseDocument import BaseDocument, BaseDocument_adder
from Products.Scheduler.Scheduler import Task


SCHEDULER_SENDMAIL_PERM = "Send notification to Newsletter"


factory_type_information = (
    {'id': 'Newsletter',
     'description': "Newsletter sends news at regular intervals.",
     'title': "_portal_type_Newsletter",
     'content_icon': 'newsletter_icon.gif',
     'product': 'NuxCPSDocuments',
     'meta_type': 'Newsletter',
     'factory': 'addNewsletter',
     'immediate_view': 'newsletter_edit_form',
     'allow_discussion': 1,
     'subscription_allowed': 1,
     'anonymous_subscription': 0,
     'allow_discussion': 1,
     'filter_content_types': 0,
     'actions': ({'id': 'view',
                  'name': '_action_view_',
                  'action': 'newsletter_view',
                  'permissions': (View,)},
                 {'id': 'edit',
                  'name': '_action_modify_',
                  'action': 'newsletter_edit_form',
                  'permissions': (ModifyPortalContent,)},
                 {'id': 'subscribe',
                  'name': '_action_(un)subscribe_',
                  'action': 'newsletter_subscribe_form',
                  'permissions': (View,)},
                 {'id': 'sendmail',
                  'name': '_action_send_email_',
                  'action': 'sendMails',
                  'permissions': (ManagePortal,)},
                 {'id': 'create',
                  'name': '_action_create_',
                  'action': 'newsletter_create_form',
                  'visible': 0,
                  'permissions': ()},
                 {'id': 'isfunctionalobject',
                  'name': 'isfunctionalobject',
                  'action': 'isfunctionalobject',
                  'visible': 0,
                  'permissions': ()},
                 ),
     },
    )


class NewsletterMailTask(Task):
    security = ClassSecurityInfo()

    security.declareProtected(View, '__init__')
    def __init__(self, newsletter, time=DateTime()):
        # Use newsletter physical id
        Task.__init__(self, newsletter.getId(), time, newsletter.absolute_url())
        self._newsletter = newsletter
    
    security.declareProtected(View, '__call__')
    def __call__(self, REQUEST=None):
        """Send email and LOGs it."""
        LOG('Newsletter' , INFO, "Task called as [%s] at [%s]"
            %(str(self), str(self.getTime())))
        # Here performs the sendMails(...) call
        LOG('Newsletter', INFO, self._newsletter.sendMails())

    def next(self):
        """Reschedule task properly."""
        next_time = None
        if(self._newsletter.frequency == 'daily'):
            next_time = self.getTime() + 1
        if(self._newsletter.frequency == 'weekly'):
            next_time = self.getTime() + 7
        if(self._newsletter.frequency == 'monthly'):
            next_time = self.getNextMonth()

        if next_time:
            new_task = NewsletterMailTask(self._newsletter, next_time)
            self._newsletter._task = new_task
            return next_time, new_task
        return None

    security.declarePublic('getNextMonthEpoch')
    def getNextMonth(self, datetime=None):
        if not datetime:
            datetime = self.getTime()
        month = datetime.month()
        year = datetime.year()
        month += 1
        if 12 < month:
            month, year = 1, year+1
        return DateTime(year, month, datetime.day())


class Newsletter(BaseDocument):
    """Newsletter"""
    meta_type = 'Newsletter'

    security = ClassSecurityInfo()

    _properties = BaseDocument._properties + (
        {'id': 'frequency', 'type': 'string', 'mode': 'w', 
         'label': 'Frequency'},
        {'id': 'subscribers', 'type': 'lines', 'mode': 'w', 
         'label': 'Subscribers'},
        {'id': 'last_sending', 'type': 'date', 'mode': 'w', 
         'label': 'Last sending'},
        {'id': 'subscription_allowed', 'type': 'boolean', 'mode': 'w', 
         'label': 'Subscription allowed'},
        {'id': 'anonymous_subscription', 'type': 'boolean', 'mode': 'w', 
         'label': 'Anonymous subscription'},
    )
    frequency = ''
    subscription_allowed = 1
    anonymous_subscription = 0
    last_sending = DateTime(0)
    max_task_age = 600

    def __init__(self, id, **kw):
        self.subscription_allowed = kw.get("subscription_allowed", 0)
        self.anonymous_subscription = kw.get("anonymous_subscription", 0)
        BaseDocument.__init__(self, id, **kw)

    security.declareProtected(ModifyPortalContent, 'edit')
    def edit(self, **kw):
        """Edit the document."""
        scheduler = getToolByName(self,"portal_scheduler")
        if hasattr(self,"_task") and self._task:
            try:
                # 1. it may have been deleted  by third party
                # 2. the scheduler is supposed to give unique time to
                #    each task
                scheduler.deschedule(self._task.getTime())
            except KeyError:
                LOG('Scheduler', ERROR,
                    "Could NOT deschedule task at time = [%s] (%s)" % (
                        str(self._task.getTime()), 
                        str(self._task)))
            del self._task

        self.subscription_allowed = kw.get("subscription_allowed", 0)
        self.anonymous_subscription = kw.get("anonymous_subscription", 0)

        # also updates self.frequency
        self.manage_changeProperties(**kw)

        if self.frequency != 'manual':
            self._task = NewsletterMailTask(self)
            scheduler.schedule(self._task.getTime(), self._task)
            LOG('Newsletter', INFO, "periodic sending task scheduled = [%s]"
                % str(self._task))
        self.reindexObject()

    #
    # Content accessors
    #
    security.declareProtected(View, 'SearchableText')
    def SearchableText(self):
        """Used by the catalog for basic full text indexing."""
        return '%s %s' % (self.title, self.description)

    #
    # Subscription
    #
    def isSubscribed(self):
        user_id = getSecurityManager().getUser().getId() 
        return user_id in self.subscribers

    def subscribe(self):
        if not self.isSubscribed():
            user_id = getSecurityManager().getUser().getId() 
            self.subscribers += (user_id,)

    def unsubscribe(self):
        if self.isSubscribed():
            user_id = getSecurityManager().getUser().getId() 
            l = [ id for id in self.subscribers if id != user_id ]
            self.subscribers = tuple(l)

    #
    # Newsletter creation
    #
    def mustSend(self):
        return 1

        # TODO: determine if this function is *dead code*
        if self.frequency == 'daily' \
          and (DateTime() - self.last_sending) > 24 * 60 * 60:
            return 1
        if self.frequency == 'weekly' \
          and (DateTime() - self.last_sending) > 7 * 24 * 60 * 60:
            return 1
        # FIXME
        if self.frequency == 'monthly' \
          and (DateTime() - self.last_sending) > 31 * 7 * 24 * 60 * 60:
            return 1

    def _getHierarchy(self):
        return self.aq_parent

    def listSubHierarchiesKeys(self):
        """Returns: the query dictionary suitable
        to list this object's Hierarchy children."""
        hierarchy = self._getHierarchy()
        if hasattr(hierarchy, "buildQuery"):
            return hierarchy.buildQuery()
        else:
            return {"PublicCategories":hierarchy.getId(),}
            

    def listObjectsToBeSent(self):
        """Returns: this object's Hierarchy children."""
        query = self.listSubHierarchiesKeys()
        if not query["PublicCategories"]:
            return []
        query['sort_on'] = 'ModificationDate'
        query['sort_order'] = 'reverse'

        l = getToolByName(self, 'portal_catalog')(query)
        return [ self.cpsdocument_info_get(brain=brain) for brain in l ]

    def sendMail(self, mail_to, mail_from, subject, content):
        if not hasattr(self, 'MailHost'):
            return 0
        mailhost = self.MailHost

        for mail_to in self.get_emails():
            if not mail_to.startswith('no_mail_for'):
                mailhost.send(content, mto=mail_to, mfrom=mail_from,
                              subject=subject, encode='8bit')

    security.declareProtected(ManagePortal, 'sendMails')
    def sendMails(self):
        """Send mails to subscribers (if it's time to do so)."""

        objects = self.listObjectsToBeSent()

        #if not self.mustSend():
        #    return 'Nothing to send'

        msg = []

        stool = getToolByName(self, "portal_subscription")
        subscribers = stool.listSubscribersFor(self)
        for subscriber in subscribers:
            if subscriber.startswith("email:"):
                subscriber_email = subscriber[6:]
            else:
                subscriber_email = self.get_member_email(subscriber)

            html = self.newsletter_mail_template(objects=objects)
            content = self.newsletter_mail(email=subscriber_email,
                subject=self.title, html=html)
            subject = self.title
            from_address = self.email_from_address

            self.sendMail(subscriber_email, self.email_from_address,
                          self.title, content)
            msg.append("Newsletter sent to " + subscriber_email)

        return msg

    #
    # Subscription
    #
    security.declareProtected(View, 'subscribe')
    def getSubscriber(self, REQUEST=None):
        """Subscribe user to Post."""
        stool = getToolByName(self, "portal_subscription")
        return stool.getMemberOrMail(REQUEST)

    security.declareProtected(View, 'isSubscribed')
    def isSubscribed(self, REQUEST=None):
        """Return the user subscription state."""
        stool = getToolByName(self, "portal_subscription")
        return stool.isSubscribed(stool.getMemberOrMail(REQUEST), self)

    security.declareProtected(View, 'unsubscribe')
    def unsubscribe(self, REQUEST=None):
        """Remove user subscription."""
        stool = getToolByName(self, "portal_subscription")
        return stool.unsubscribe(stool.getMemberOrMail(REQUEST), self)

    security.declareProtected(View, 'subscribe')
    def subscribe(self, REQUEST=None):
        """Subscribe user to Post."""
        stool = getToolByName(self, "portal_subscription")
        return stool.subscribe(stool.getMemberOrMail(REQUEST), self)

    security.declareProtected(View, 'listSubscriptions')
    def listSubscriptions(self, REQUEST=None):
        """Return observed objects for current user."""
        stool = getToolByName(self, "portal_subscription")
        return [ object for object 
                 in stool.listObservedObjectsFor(stool.getMemberOrMail(REQUEST))
                 if object.portal_type == self.portal_type ]

    #
    #
    # Behavior
    #
    security.declarePrivate('get_user_folder')
    def get_user_folder(self):
        return self.acl_users

    security.declarePrivate('get_member_email')
    def get_member_email(self, id):
        dtool = getToolByName(self, 'portal_metadirectories')
        entry = dtool.members.getEntry(id)
        if not entry:
            return ''
        email = entry.get('email')
        if not email: # XXX
            email = 'no_mail_for_%s@example.com' % id
        return email

    security.declarePrivate('get_group_emails')
    def get_group_emails(self, gid):
        # XXX should use portal_directory somehow
        aclu = self.get_user_folder()
        if not hasattr(aclu, 'getGroupById'):
            return []
        group = aclu.getGroupById(gid, None)
        if group is None:
            return []
        users = group.getUsers()
        emails = []
        for id in users:
            email = self.get_member_email(id)
            if email:
                emails.append(email)
        return emails

    security.declareProtected(View, 'get_emails')
    def get_emails(self, context=None):
        """Get the emails for this list

        If context is not None, use it instead of self
        as the place to lookup workgroup managers/members/visitors.
        """
        ids = []
        # wg managers, members or visitors
        if context is None:
            context = aq_parent(self)
        
        getContextUsers = getattr(context, 'getContextUsers', None)

        if getContextUsers is not None:
            context_users = getContextUsers()
        else:
            context_users = None
            mlr = self.mergedLocalRoles(context)

        for var, role in (('add_wg_managers', 'WorkgroupManager'),
                          ('add_wg_members', 'WorkgroupMember'),
                          ('add_wg_visitors', 'WorkgroupVisitor')):
            if not hasattr(self, var):
                continue
            if context_users is None:
                for id, roles in mlr.items():
                    if role in roles:
                        ids.append(id)
            else:
                for id in context_users.get(role, []):
                    ids.append(id)

        # treat ids
        emails = []
        for id in ids:
            if id.startswith('group:'):
                gid = id[len('group:'):]
                emails.extend(self.get_group_emails(gid))
            else:
                email = self.get_member_email(id)
                if email:
                    emails.append(email)

        # add explicit emails
        stool = getToolByName(self, "portal_subscription")
        for subscriber in stool.listSubscribersFor(self):
            if subscriber.startswith("email:"):
                emails.append(subscriber[6:])
            else:
                emails.append(self.get_member_email(subscriber))

        # remove redundant or invalid emails
        okemails = []
        for email in emails:
            email = email.strip()
            if email.find('@') <= 0:
                continue
            if email not in okemails:
                okemails.append(email)

        return okemails


    security.declareProtected(View, 'send_email')
    def send_email(self,
                   from_address='nobody@example.com',
                   reply_to=None,
                   subject='No subject',
                   body='No body',
                   context=None,
                   content_type='text/plain',
                   charset='ISO-8859-15',
                   mto=None):
        """Send a mail to the list

        If context is not None, use it instead of self as the place to lookup
        workgroup managers/members/visitors.
        """
        if not hasattr(self, 'MailHost'):
            return 0
        mailhost = self.MailHost
        if mto is None:
            mto = self.get_emails(context=context)
            mto = [email for email in mto 
                   if not email.startswith('no_mail_for')]
        if reply_to is None:
            reply_to = from_address
        content = """\
From: %s
To: %s
Reply-To: %s
Subject: %s
Content-Type: %s; charset=%s
Mime-Version: 1.0

%s"""
        content = content % (
            from_address, reply_to, ', '.join(mto), subject, 
            content_type, charset, body)
        mailhost.send(content, mto=mto, mfrom=from_address,
                      subject=subject, encode='8bit')


    # Modified from CMFCore.utils
    # XXX should be moved to NuxUserGroups
    security.declarePrivate('mergedLocalRoles')
    def mergedLocalRoles(self, object):
        """
        Return a merging of object and its ancestors' __ac_local_roles__
        and __ac_local_group_roles__
        """
        merged = {}
        object = getattr(object, 'aq_inner', object)
        while 1:
            if hasattr(object, '__ac_local_roles__'):
                dict = object.__ac_local_roles__ or {}
                if callable(dict): dict = dict()
                for k, v in dict.items():
                    if merged.has_key(k):
                        merged[k] = merged[k] + v
                    else:
                        merged[k] = v
            # groups
            if hasattr(object, '__ac_local_group_roles__'):
                dict = object.__ac_local_group_roles__ or {}
                if callable(dict): dict = dict()
                for k, v in dict.items():
                    k = 'group:'+k
                    if merged.has_key(k):
                        merged[k] = merged[k] + v
                    else:
                        merged[k] = v
            # end groups
            if hasattr(object, 'aq_parent'):
                object = object.aq_parent
                object = getattr(object, 'aq_inner', object)
                continue
            if hasattr(object, 'im_self'):
                object = object.im_self
                object = getattr(object, 'aq_inner', object)
                continue
            break
        return merged


InitializeClass(Newsletter)

def addNewsletter(dispatcher, id, REQUEST=None, **kw):
    """Add a Newsletter."""
    ob = Newsletter(id, **kw)
    return BaseDocument_adder(dispatcher, id, ob, REQUEST=REQUEST)

