#
# Copyright 2001 Stfane Fermigier and Nuxeo SARL
# See LICENSE.TXT for licensing information
#

"""NuxDocument: a Zope product that represents generic documents by using
plugins to convert native productivity suite formats to HTML (for viewing or
previewing in a regular browser) or to plain text (for indexing).
"""

__version__='$Revision: 1.47 $'[11:-2]

from zLOG import LOG, DEBUG, BLATHER, ERROR
from string import maketrans
from OFS.SimpleItem import SimpleItem
from OFS.ObjectManager import ObjectManager
from OFS.Cache import Cacheable
from OFS.Image import File
from DateTime.DateTime import DateTime
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass

import plugins

try:
    from Products.NuxStorage.FileProxy import FileProxy
    from Products.CMFCore.utils import getToolByName
    LOG('NuxDocument', BLATHER, 'PortalStorage support enabled')
    USE_PORTAL_STORAGE = 1
except ImportError:
    LOG('NuxDocument', BLATHER, 'PortalStorage not found, support disabled')
    USE_PORTAL_STORAGE = 0


class NuxDocument(SimpleItem, Cacheable, ObjectManager):
    # (SimpleItem):
    # (CatalogAware, SimpleItem, Acquisition.Implicit, 
    # PropertyManager, Historical, Cacheable, Item_w__name__):
    # + ElementWithTitle (XXX ???)

    """The NuxDocument base class. Provides only state and behavior,
    UI in provided by subclasses."""

    index_html = None # Prevent accidental acquisition (XXX ???)

    filename = _text = _html = _mime_type = ''
    _file = None
    _properties = (
        {'id': 'filename', 'type': 'string', 'mode': 'r'},
        #{'id': 'mime_type', 'type': 'string', 'mode': 'r'},
        #{'id': 'author', 'type': 'string', 'mode': 'w'},
        #{'id': 'publisher', 'type': 'string', 'mode': 'w'},
        #{'id': 'source', 'type': 'string', 'mode': 'w'},
        #{'id': 'publication_date', 'type': 'date', 'mode': 'w'},
        #{'id': 'description', 'type': 'text', 'mode': 'w'},
        #{'id': 'keywords', 'type': 'text', 'mode': 'w'},
    )

    security = ClassSecurityInfo()

    #
    # Constructor
    #
    def __init__(self, id='', title='', file='', mime_type=''):
        self._setId(id)
        self._edit(title, file, mime_type)


    # taken from CMFDefault.File
    def _isNotEmpty(self, file):
        """Do various checks on 'file' to try to determine non-emptiness"""
        if not file:
            return 0                    # Catches None, Missing.Value, ''
        elif file and (type(file) is type('')):
            return 1
        elif getattr(file, 'filename', None):
            return 1
        elif not hasattr(file, 'read'):
            return 0
        else:
            file.seek(0, 2)             # 0 bytes back from end of file
            t = file.tell()             # Report the location
            file.seek(0)                # and return pointer back to 0
            if t: return 1
            else: return 0

    def _edit(self, title=None, file=None, mime_type=None):
        # XXX: to be tested under Windows.
        # XXX: now filename == id (if id == cooked from filename), so ?
        filename = getattr(file, 'filename', self.__name__)
        self.filename = filename[max(
            filename.rfind('/'),
            filename.rfind('\\'),
            filename.rfind(':'),) + 1:]
        if title is not None:
            self.title = title
        notempty = self._isNotEmpty(file)
        if notempty:
            # Check if there is a portal_storage tool.
            # If not, use a standard file.
            if USE_PORTAL_STORAGE:
                pstool = getToolByName(self, 'portal_storage', None)
                if pstool is None:
                    LOG('NuxDocument', DEBUG, 'portal_storage not found',
                    str(pstool))
            else:
                pstool = None

            if pstool is None:
                self._file = File(self.__name__, title, file)
            else:
                self._file = FileProxy(self.filename, title)
                self._file.update_data(file, None, None)

        # set a default mime_type that convert() can override
        user_mime_type = ''
        if mime_type:
            user_mime_type = mime_type
            self._mime_type = mime_type
        elif notempty:
            mime_type = self._file.content_type
            if mime_type not in ('', 'application/octet-stream',
              'application/unknown'):
                user_mime_type = mime_type
            self._mime_type = mime_type
        if notempty:
            self.convert()
            # if user/browser set the mime type, keep it
            if user_mime_type:
                self._mime_type = user_mime_type
        self.publication_date = DateTime()

    def convert(self):
        data = self.getRaw()
        plugin = plugins.selectPlugin(data, mime_type=self.MimeType(),
            filename=self.filename)

        if plugin != plugins.Dumb:
            try:
                result = plugin.getConverter(data)
            except plugins.ConversionError:
                plugin = plugins.getDefaultPlugin()
                result = plugin.getConverter(data)
        else:
            result = plugin.getConverter(data)

        self._mime_type = result.getMimeType() or self.MimeType()
        self._text = result.getText()
        self._html = result.getBody()
        if not self.title:
            self.title = result.getTitle()
        for image_name in result.getImageNames():
            if hasattr(self, image_name):
                self.manage_delObjects(ids=[image_name])
            self.manage_addImage(image_name, result.getImage(image_name))

    #
    # Accessors / rendering
    #
    security.declarePrivate('getRaw')
    def getRaw(self):
        "Get raw content of document"
        file = self._file
        if file is None:
            return ''
        else:
            return str(file)

    security.declareProtected('View', 'getHtml')
    def getHtml(self):
        "Get raw HTML converted content"
        return self._html

    security.declareProtected('View', 'getHtml')
    def getText(self):
        "Get raw text converted content"
        return self._text

    security.declareProtected('View', 'getFileSize')
    def getFileSize(self):
        "Get raw content size in bytes"
        return len(self.getRaw())

    security.declareProtected('View', 'getFileSizeString')
    def getFileSizeString(self):
        """Get raw content size as a dict

        E.g.: { 'size': size, 'unit': 'B'/'KB'/'MB' }, formats size in B/KB/MB
        """
        size = self.getFileSize()
        if size < 1024:
            return { 'size': size, 'unit': 'B' }
        elif size < 1048576:
            return { 'size': size/1024, 'unit': 'KB' }
        else:
            return { 'size': size/1048576, 'unit': 'MB' }

    security.declareProtected('View', 'MimeType')
    def MimeType(self):
        """Get MIME type"""
        return self._mime_type or 'application/octet-stream'

    security.declareProtected('View', 'content_type')
    content_type = MimeType

    security.declareProtected('View', 'document_src')
    def document_src(self, REQUEST=None, RESPONSE=None):
        """Get unprocessed document source"""
        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', self.MimeType())
            filename = sanitize_filename(self.filename or '')
            if not filename:
                filename = sanitize_filename(self.getId())
                if not filename:
                    filename = 'document'
            RESPONSE.setHeader('Content-Disposition',
                               "inline; filename=%s" % filename)
            # Performance fix. Was: return self.getRaw()
            r = self.getRaw()
            RESPONSE.setHeader('Content-Length', len(r))
            RESPONSE.write(r)
            return ''
        else:
            return self.getRaw()

    security.declareProtected('FTP access', 'manage_FTPget')
    def manage_FTPget(self, **kw):
        "Get source for FTP download"
        return self.getRaw()

    security.declareProtected('View', 'HEAD')
    def HEAD(self, REQUEST, RESPONSE):
        "Return HTTP HEAD result (override webdav.Collection behavior)"
        return self._file.HEAD(REQUEST, RESPONSE)

    security.declareProtected('View', 'PrincipiaSearchSource')
    def PrincipiaSearchSource(self):
        return self.getText()

    security.declareProtected('View', 'text_content')
    text_content = PrincipiaSearchSource

    security.declareProtected('Edit NuxDocument', 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        "Handle HTTP/FTP PUT requests."
        self.dav__init(REQUEST, RESPONSE)
        file = REQUEST.get('BODY', '')
        mime_type = REQUEST.get('CONTENT_TYPE', None)
        self._edit(file=file, mime_type=mime_type)
        RESPONSE.setStatus(204)
        return ''

    def manage_beforeDelete(self, item, container):
        LOG('NuxDocument', DEBUG, 'manage_beforeDelete')
        ObjectManager.manage_beforeDelete(self, item, container)
        if self._file is not None and hasattr(self._file, 'manage_beforeDelete'):
            self._file.manage_beforeDelete(item, container)

    def manage_afterClone(self, item):
        LOG('NuxDocument', DEBUG, 'manage_afterClone')
        ObjectManager.manage_afterClone(self, item)
        if self._file is not None and hasattr(self._file, 'manage_afterClone'):
            self._file.manage_afterClone(item)

    def manage_afterAdd(self, item, container):
        LOG('NuxDocument', DEBUG, 'manage_afterAdd')
        ObjectManager.manage_afterAdd(self, item, container)
        if self._file is not None and hasattr(self._file, 'manage_afterAdd'):
            self._file.manage_afterAdd(item, container)

InitializeClass(NuxDocument)


def sanitize_filename(filename):
    filename = filename or ''
    filename = filename[max(filename.rfind('/'),
                            filename.rfind('\\'),
                            filename.rfind(':'))+1:]
    filename = filename.replace('', 'AE')
    filename = filename.replace('', 'ae')
    # Losing accents may not be necessary
    tr = maketrans('',
                   'AAAAAACEEEEIIIINOOOOOOUUUUYaaaaaaceeeeiiiinoooooouuuuyy')
    filename = filename.translate(tr)
    ok = ('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
          ' -_.,!?()[]&%$' "'") # at least not = ; / \ :
    filename = ''.join([c for c in filename if c in ok])
    filename = filename.strip()
    return filename

