# Copyright (c) 2002 Nuxeo SARL <http://nuxeo.com>
# Author: Florent Guillaume <mailto:fg@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: WorkflowManager.py,v 1.5 2003/07/14 21:18:14 fermigier Exp $

"""
  WorkflowManager Mixin
"""

__version__ = '$Revision: 1.5 $'[11:-2]

import string

from zLOG import LOG, DEBUG

from Acquisition import aq_base
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile
from DateTime.DateTime import DateTime

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.CMFCorePermissions import \
     setDefaultRoles, View, ModifyPortalContent
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFCore.PortalFolder import PortalFolder
from Products.CMFCore.PortalContent import PortalContent


from WorkflowAware import WorkflowAware


class WorkflowManager(WorkflowAware):
    """
    WorkflowManager Mixin.
    Give Object Managers the capability to deny their children
    participation in some workflows.

    Must be at the front of the inheritance chain because it
    redefines manage_afterAdd and manage_beforeDelete.
    Base class must also derive from PortalContent and PortalFolder.
    """

    _isWorkflowManager = 1

    _workflow_children_slavedchain = () # Slaved workflow chain

    #manage_options = (
    #            {'label':'WorkflowManager', 'action':'manage_workflowManager',
    #             'help':('OFSP','Folder_View.stx')},
    #            )

    security = ClassSecurityInfo()

    #
    # Workflow Manager accessors
    #

    security.declarePrivate('setWorkflowManagerSlavedChain')
    def setWorkflowManagerSlavedChain(self, chain=()):
        """
        Set the children slaved workflow chain.
        """
        self._workflow_children_slavedchain = chain

    #
    # Workflow Manager hooks
    #

    security.declarePrivate('workflowManagerGetChain')
    def workflowManagerGetChain(self, ob, chain, pwf=None):
        """
        Hook called by Workflow Tool after computing the chain.
        Return the workflow chain to use for a child.
        """
        slavedchain = self._workflow_children_slavedchain
        for id in slavedchain:
            if id in chain:
                chain.remove(id)
        return chain

    #def workflowManagerListActions(self, info, actions, pwf=None):
    # obsolete

    security.declarePrivate('workflowManagerDoAction')
    def workflowManagerDoAction(self, ob, action, wf_id, args, kw, pwf=None):
        """
        Hook called by Workflow Tool before calling wf.doActionFor.
        Ensure the action is in one of the allowed workflows.
        """
        if wf_id is not None:
            slavedchain = self._workflow_children_slavedchain
            if wf_id in slavedchain:
                # if explicit wf_id in slavedchain, error
                raise WorkflowException(
                    'Slaved child has no workflow action "%s".' % action)
        # continue with normal doActionFor which will ignore slaved workflows
        return

    # XXX workflow notifyCreated shouldn't reset the workflow
    # after a copy, it should call notifyCopied or something

    security.declarePrivate('workflowManagerNotifyCreated')
    def workflowManagerNotifyCreated(self, ob, pwf=None):
        """
        Hook called by Workflow Tool before calling wf.notifyCreated.
        Notify the Workflow Manager that an object has been created/moved.
        """
        # note that by default the slaved workflows would not be notified
        slavedchain = self._workflow_children_slavedchain
        # decide in what state to put the object in the slaved workflows
        if pwf is None:
            pwf = getToolByName(self, 'portal_workflow')
        for id in slavedchain:
            wf = pwf.getWorkflowById(id)
            if wf is not None:
                wf.notifyCreated(ob)
                # The workflow may provide options to put new objects
                # in a state similar to the Workflow Manager.

    #
    # Workflow Aware hooks
    #

    # override
    def workflowDoAction(self, action, wf_id, args, kw, pwf=None):
        """
        Recurse in the children if the action is a slaved one.
        """
        ob = self
        if pwf is None:
            pwf = getToolByName(self, 'portal_workflow')
        # find wf_id
        if wf_id is None:
            wfs = pwf.getWorkflowsFor(ob)
            if not wfs:
                return
            for wf in wfs:
                if wf.isActionSupported(ob, action):
                    wf_id = wf.getId()
                    break
        # if action's workflow is in slaved, do recursion in the children
        slavedchain = self._workflow_children_slavedchain
        if wf_id is not None and wf_id in slavedchain:
            LOG('WorkflowManager', DEBUG, ' children slaved')
            self._recursiveDoAction(ob, action, wf_id, args, kw, pwf)
        return

    security.declarePrivate('_recursiveDoAction')
    def _recursiveDoAction(self, ob, action, wf_id, args, kw, pwf):
        """
        Recurse in the children except inside those workflow aware.
        """
        LOG('WorkflowManager', DEBUG, ' in _recursiveDoAction, obid=%s'
            % ob.getId())
        if hasattr(aq_base(ob), 'objectValues'):
            subobjs = ob.objectValues()
            for subobj in subobjs:
                LOG('WorkflowManager', DEBUG, ' doing subobjid=%s'
                    % subobj.getId())
                # XXX check first that the object is really inside the wf_id?
                #try:
                apply(pwf.doActionFor_NoWFM, (subobj, action, wf_id)+args, kw)
                #except WorkflowException:
                #    pass # why exactly ?
                if not getattr(subobj, '_isWorkflowAware', 0):
                    self._recursiveDoAction(subobj, action, wf_id, args, kw,
                        pwf)

    #
    # AfterAdd and BeforeDelete are both contentish and folderish
    #

    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        """
        Add self to the workflow and catalog.
        """
        # Recurse in the children (ObjectManager)
        PortalFolder.manage_afterAdd(self, item, container)
        # Add self to workflow and catalog
        PortalContent.manage_afterAdd(self, item, container)

    security.declarePrivate('manage_beforeDelete')
    def manage_beforeDelete(self, item, container):
        """
        Remove self from the workflow and catalog.
        """
        # We don't call PortalContent.manage_beforeDelete because
        # it deals with talkbacks which are subobjects...
        # Remove self from catalog
        if aq_base(container) is not aq_base(self):
            self.unindexObject()
        # Then recurse in the children (ObjectManager)
        PortalFolder.manage_beforeDelete(self, item, container)

    # manage_afterClone ?
    # here we would update unique identifiers


InitializeClass(WorkflowManager)
