# Copyright (c) 2002 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2002 Prfecture du Bas-Rhin, France
# Author: Florent Guillaume <mailto:fg@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: BaseHierarchyLevel.py,v 1.13 2003/07/14 21:30:14 fermigier Exp $

"""
  Base Hierarchy Level
"""

__version__ = '$Revision: 1.13 $'[11:-2]

from zLOG import LOG, DEBUG
import string
from random import randrange

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base

from Products.ZCatalog.ZCatalog import ZCatalog

from Products.CMFCore.CMFCorePermissions import AccessContentsInformation
from Products.CMFCore.utils import getToolByName
from Products.PortalContentFolder.PortalContentFolder import PortalContentFolder
from DateTime.DateTime import DateTime

from HierarchiesTool import RebuildHierarchy, ChangeHierarchyLevel


BadRequest = 'Bad Request'


class BaseHierarchyLevel(PortalContentFolder):
    """
    Base class for a Hierarchy Level
    """
    meta_type = None
    isPortalContent = 0
    isDocumentContainer = 0 # to distinguish folder view method from workgroups

    _properties = (
        {'id':'title', 'type':'string', 'mode':'w',
            'label':'title'},
        {'id':'description', 'type':'text', 'mode':'w',
            'label':'description'},
        {'id':'hierarchyid', 'type':'string', 'mode':'w',
            'label':'Hierarchy name'},
        {'id':'levelkey', 'type':'string', 'mode':'w',
            'label':'Level key'},
        {'id':'publishable', 'type':'boolean', 'mode':'w',
            'label':'Publication allowed'},
        {'id':'skinnerdocument', 'type':'string', 'mode':'w',
            'label':'Skinner'},
        {'id':'hierarchy_order', 'type':'lines', 'mode':'w',
            'label':'Publication order'},
        )

    title = ''
    description = ''
    hierarchyid = 'main'
    levelkey = ''
    publishable = 1
    skinnerdocument = ''
    hierarchy_order = []

    security = ClassSecurityInfo()

    def __init__(self, id,
                 title='',
                 description='',
                 hierarchyid='main',
                 levelkey='',
                 publishable=1,
                 hierarchy_order=[],
                 **kw):
        PortalContentFolder.__init__(self, id, title=title)
        self.description = description
        self.hierarchyid = hierarchyid
        if not levelkey:
            levelkey = str(int(DateTime())) + str(randrange(1000, 10000))
        self.levelkey = levelkey
        self.publishable = publishable
        self.hierarchy_order = hierarchy_order

    #
    # Misc
    #

    security.declarePublic('propertyLabel')
    def propertyLabel(self, id):
        """Return a label for the given property id."""
        for p in self._properties:
            if p['id'] == id:
                return p.get('label', id)
        return id

    #
    # Accessors
    #

    security.declarePublic('getHierarchyId')
    def getHierarchyId(self):
        """Get the Hierarchy Id to which this level belongs."""
        return self.hierarchyid

    security.declarePublic('getLevelKey')
    def getLevelKey(self):
        """Get the Level Key for this level."""
        return self.levelkey

    security.declarePublic('getPublishable')
    def getPublishable(self):
        """Get the Publishable flag for this level."""
        return self.publishable

    security.declarePublic('publicationAllowedForUser')
    def publicationAllowedForUser(self, id=None):
        """Return true if the user (current=None) can publish to this level."""
        if not self.publishable:
            return 0
        if id is None:
            pm = getToolByName(self, 'portal_membership')
            member = pm.getAuthenticatedMember()
            if not hasattr(member, 'getMemberId'):
                return 0
            id = member.getMemberId()
        return 1

    
    security.declarePublic('getHierarchyOrder')
    def getHierarchyOrder(self):
        """Return the ordered list of publication"""
        return self.hierarchy_order
    
    #
    # Helpers methods
    #

    security.declareProtected(RebuildHierarchy, 'rebuildHierarchy')
    def rebuildHierarchy(self):
        """Rebuilds the hierarchy."""
        htool = getToolByName(self, 'portal_hierarchies')
        if htool is not None:
            htool.getHierarchyById(self.getHierarchyId()).rebuildHierarchy()

    #
    # Edition
    #

    security.declareProtected(ChangeHierarchyLevel, 'edit')
    def edit(self,
             title=None,
             description=None,
             hierarchyid=None,
             levelkey=None,
             publishable=None,
             hierarchy_order=None,
             REQUEST=None, **kw):
        """
        Edits a Hierarchy Level.
        """
        if title is not None:
            self.title = title
        if description is not None:
            self.description = description
        if levelkey is not None:
            if not levelkey:
                levelkey = str(int(DateTime())) + str(randrange(1000, 10000))
            self.levelkey = levelkey
        if hierarchyid is not None:
            self.hierarchyid = hierarchyid
        if publishable is not None:
            self.publishable = publishable
        if hierarchy_order is not None:
            self.hierarchy_order = hierarchy_order

    #
    # Hierarchy methods
    #

    security.declarePrivate('get_level_additional_info')
    def get_level_additional_info(self, dict):
        """Add to dict additional info."""
        pass

    def _filter_samehier(self, obj):
        # Get all the children items of obj
        # with the same hierarchyid and meta_type as self
        items = obj.objectItems([self.meta_type])
        hierarchyid = self.hierarchyid
        samehier = lambda item, hid=hierarchyid: item[1].hierarchyid == hid
        res = filter(samehier, items)
        return res

    security.declareProtected(AccessContentsInformation, 'getChildrenItems')
    def getChildrenItems(self):
        """Get the list of children items."""
        return self._filter_samehier(self)

    security.declareProtected(AccessContentsInformation, 'getChildrenValues')
    def getChildrenValues(self):
        """Get the list of children values."""
        items = self.getChildrenItems()
        second = lambda x: x[1]
        return map(second, items)

    security.declareProtected(AccessContentsInformation, 'getSiblingsItems')
    def getSiblingsItems(self):
        """Get the list of siblings items."""
        parent = self.aq_inner.aq_parent
        return self._filter_samehier(parent)

    security.declareProtected(AccessContentsInformation, 'getSiblingsValues')
    def getSiblingsValues(self):
        """Get the list of sliblings values."""
        items = self.getSiblingsItems()
        second = lambda x: x[1]
        return map(second, items)

    security.declareProtected(AccessContentsInformation, 'getParentItems')
    def getParentItems(self):
        """Get the parent items or ()."""
        parent = self.aq_inner.aq_parent
        if parent.meta_type == self.meta_type and \
           parent.hierarchyid == self.hierarchyid:
            return [(parent.id, parent)]
        else:
            return ()

    security.declareProtected(AccessContentsInformation, 'getParentValues')
    def getParentValues(self):
        """Get the list of parent values."""
        items = self.getParentItems()
        second = lambda x: x[1]
        return map(second, items)

    #
    # Query methods
    #

    def _get_hierarchy(self):
        # Get the hierarchy object to which this level belongs.
        portal_hierarchies = getToolByName(self, 'portal_hierarchies', None)
        if portal_hierarchies is None:
            raise BadRequest, 'No Hierarchies Tool has been defined.'
        return portal_hierarchies.getHierarchyById(self.hierarchyid)

    security.declareProtected(AccessContentsInformation, 'buildQuery')
    def buildQuery(self):
        """
        Builds the query for this level.
        """
        levelkey = self.levelkey
        if not levelkey:
            return {}
        criterion = self._get_hierarchy().getCriterion()
        return {criterion: levelkey}

    security.declareProtected(AccessContentsInformation, 'queryCatalog')
    def queryCatalog(self, REQUEST=None, **kw):
        """
        Queries the Catalog to get this level's documents.
        """
        query = self.buildQuery()
        if not query:
            return ()
        kw.update(query)
        portal_catalog = getToolByName(self, 'portal_catalog')
        return portal_catalog.searchResults(REQUEST, **kw)

    security.declarePrivate('queryCatalogNoSecurity')
    def queryCatalogNoSecurity(self, REQUEST=None, **kw):
        """
        Queries the Catalog to get this level's documents.
        Accesses the ZCatalog directly to bypass allowedRolesAndUsers
        filtering. This is needed to do a query during traversal where
        the user is not yet authenticated.
        """
        query = self.buildQuery()
        if not query:
            return ()
        kw.update(query)
        portal_catalog = getToolByName(self, 'portal_catalog')
        return ZCatalog.searchResults(portal_catalog, REQUEST, **kw)

    # XXX unused (unimplemented in hier)
    security.declareProtected(AccessContentsInformation, 'buildRecursiveQuery')
    def buildRecursiveQuery(self):
        """Build the query for this level and all sublevels."""
        return self._get_hierarchy().buildRecursiveQuery(self)

    # XXX unused (unimplemented in hier)
    security.declareProtected(AccessContentsInformation, 'queryCatalogRecursive')
    def queryCatalogRecursive(self, REQUEST=None, **kw):
        """Query the Catalog to get this level's and all sublevel's data."""
        kw.update(self.buildRecursiveQuery())
        portal_catalog = getToolByName(self, 'portal_catalog')
        return apply(portal_catalog.searchResults, (REQUEST,), kw)

    def manage_afterAdd(self, item, container):
        if aq_base(item) is aq_base(self):
            # rebuild hierarchy
            self.rebuildHierarchy()
        PortalContentFolder.manage_afterAdd(self, item, container)

    def manage_afterClone(self, item):
        PortalContentFolder.manage_afterClone(self, item)
        # change hierarchy id and rebuild hierarchy
        self.edit(levelkey='')
        self.rebuildHierarchy()

    #
    # Traversal
    #

    def __getitem__(self, name):
        """
        Gets to the virtual documents as if they were children.
        """
        # WARNING when called during traversal, the current user
        # is still Anonymous so we do query directly the ZCatalog
        # to bypass the allowedRolesAndUsers check of Catalog Tool.
        # There are no security implications because the final object
        # will still be checked for security by the Publisher.

        # XXX id maybe not unique... use something unique here
        res = self.queryCatalogNoSecurity(id=name)
        nb = len(res)
        if nb != 1:
            if hasattr(aq_base(self), 'queryPendingCatalogNoSecurity'):
                res = self.queryPendingCatalogNoSecurity(id=name)
                nb = len(res)
        if nb != 1:
            raise KeyError, name
        target = res[0].getObject()
        try:
            reqo = self.REQUEST.other
        except AttributeError:
            reqo = None
        if reqo is not None:
            realpath = target.getPhysicalPath()
            virtualpath = self.getPhysicalPath() + (name,)
            info = (realpath, virtualpath)
            reqo['munge_absolute_url'] = info
        # Return the document
        return target

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

InitializeClass(BaseHierarchyLevel)

