##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
# 
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
# 
##############################################################################
"""
Scheduler module.

$Id: Scheduler.py,v 1.1 2003/03/12 11:02:29 smigniot Exp $
"""
import os, time, sys
from popen2 import Popen3
from types import FloatType, IntType

import Globals
from BTrees import IOBTree
from Persistence import Persistent
from ExtensionClass import Base
from Acquisition import Implicit
from Acquisition import aq_base
from AccessControl import ClassSecurityInfo

from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from ZPublisher.mapply import mapply
from ZPublisher.Publish import call_object, missing_name, dont_publish_class

from Products.CMFCore.CMFCorePermissions import View, ManagePortal
from Products.PageTemplates.PageTemplate import PageTemplate
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.Event.ISubscriptionAware import ISubscriptionAware
from Products.Event.ISubscriber import ISubscriber

from Products.Scheduler import IScheduler
from Products.Scheduler import IScheduledEvent
from Products.Scheduler import IDescheduledEvent
from Products.Scheduler import ITimeEvent
from SchedulerPermissions import *

from zLOG import LOG, PROBLEM, ERROR, BLATHER, DEBUG, INFO
import xmlrpclib

from signal import SIGTERM

__version__ = "$Revision: 1.1 $"[11:-2]

class Scheduler(SimpleItem, PropertyManager):
    meta_type = 'Scheduler'

    manage_options = (
        {'label': 'Current Tasks', 'action': 'manage_current_tasks'},
        {'label': 'Schedule A Task', 'action': 'manage_schedule_tasks'},
        {'label': 'Properties', 'action':'manage_propertiesForm'},
        {'label': 'Clock', 'action':'manage_clock'},
        {'label': 'Security', 'action':'manage_access'},
        {'label': 'Ownership', 'action':'manage_owner'},
        {'label': 'Undo', 'action':'manage_UndoForm'},
        )
    _properties = (
        {'id':'steps', 'type':'lines', 'mode':'w'},
        {'id':'event_service', 'type':'string', 'mode':'w'},
        {'id':'filter_data', 'type':'string', 'mode':'r'},
        {'id':'interval', 'type':'int', 'mode':'w'},
        {'id':'login', 'type':'string', 'mode':'w'},
        {'id':'password', 'type':'string', 'mode':'w'},
        )
    
    # offering to the security gods
    security = ClassSecurityInfo()
    ok = {'meta_type':1, 'id':1, 'icon':1, 'bobobase_modification_time':1 ,
          'title_or_id': 1, 'filter_data': 1, 'event_service': 1}

    security.setDefaultAccess(ok)
    security.setPermissionDefault(MGMT_SCREEN_PERM, ['Manager'])
    security.setPermissionDefault(CHANGE_SCHEDULE_PERM, ['Manager'])
    security.setPermissionDefault(VIEW_SCHEDULE_PERM, ['Manager'])
    security.setPermissionDefault(NOTIFY_SCHEDULE_PERM, ['Manager'])
        
    __implements__ = (IScheduler, ISubscriptionAware, ISubscriber)

    def __init__(self, id, title='', filter_data='',
                 event_service='portal_events'):
        self.id = id
        self.title = title
        self.subscribed = []
        self.tasks = IOBTree.IOBTree()
        if filter_data:
            self.filter_data = filter_data
        else:
            self.filter_data = ''
        self.event_service = event_service
        self.steps = []
        self.interval = 0
        self.login = "anonymous"
        self.password = "nopass"

    def subscribedTo(self, subscribable, event_type, filter):
        """
        alerts the object that it has subscribed, via a call from
        itself or from another object, to the subscribable.  The
        event_type and filter match the arguments provided to
        ISubscribable.subscribe.
        """
        subscribable = getattr(subscribable, 'aq_base', subscribable)
        self.subscribed.append((subscribable, event_type, filter))
        self._p_changed = 1
    
    def unsubscribedFrom(self, subscribable, event_type, filter):
        """alerts the object that it has unsubscribed, via a call from
        itself or from another object, to the subscribable.  The
        event_type and filter match the exact event_type and filter of
        the deleted subscription, rather than, necessarily, the
        arguments provided to ISubscribable.unsubscribe.
        """
        subscribable = getattr(subscribable, 'aq_base', subscribable)
        self.subscribed.remove((subscribable, event_type, filter))
        self._p_changed = 1

    def getFilter(self):
        if self.filter_data:
            return Filter(self.filter_data)
        return None

    def manage_afterAdd(self, item, container):
        event_service = getattr(self, self.event_service, None)
        if event_service is not None:
            # receive all time events
            event_service.subscribe(self, ITimeEvent)
            # receive schedule events that match our id
            event_service.subscribe(self, IScheduledEvent,
                                    filter=self.getFilter())

    def manage_beforeDelete(self, item, container):
        event_service = getattr(self, self.event_service, None)
        if event_service is not None:
            # un-receive all time events
            event_service.unsubscribe(self, ITimeEvent)
            # un-receive schedule events that match our id
            event_service.unsubscribe(self, IScheduledEvent,
                                      filter=self.getFilter())

    security.declareProtected(NOTIFY_SCHEDULE_PERM, 'notify')
    def notify(self, event=None):
        """
        If it is specified, 'event' must be a "time event" or a
        "schedule event".

        A time event may be signified by any of the following objects:
        
          - a float or an int (although the ability for the event to be
            a float or int is not specified in the IScheduler interface)

          - None, which is translated into a time event represented by the
            current time.

          - an object that implements the ITimeEvent interface

        When we receive a time event, we process all events in the queue
        that have a scheduled time before the time event.

        A schedule event is an object that implements the IScheduledEvent
        interface.  When we receive a schedule event, we schedule the
        event but we do not actually process the event queue.
        """
        LOG('Scheduler', BLATHER, 'notify called')
        if event is None:
            # If event is None, assume that we want to run tasks before
            # "now" by turning event into a time.time as of now.
            # This is the common case.
            t = int(time.time())
        elif (isinstance(event, IntType) or isinstance(event, FloatType)):
            # Check if event is an integer or a float.
            t = int(event)
        elif ITimeEvent.isImplementedBy(event):
            # Check if event is an object that implements the ITimeEvent
            # interface that we need to get a number time from
            t = int(event.getTime())
        elif IScheduledEvent.isImplementedBy(event):
            # this is a schedule event, we want to schedule this event
            # with ourselves, but we don't want to actually perform
            # any tasks.
            self.schedule(event.getTime(), event)
            return
        elif IDescheduledEvent.isImplementedBy(event):
            # this is a deschedule event, we want to deschedule this event,
            # but we don't want to actually perform any tasks.
            self.deschedule(event.getTime())
            return
        else:
            # We don't know what to do, so we punt.
            raise ValueError, (
                '"event" must be an object that supports the ITimeEvent '
                'interface, an object that supports the IScheduledEvent '
                'interface, an integer, or a float.  %s is not a valid '
                'value.' % event)

        for this_time, this_task in self.getPendingTasks(t):
            self.deschedule(this_time) # deschedule the task
            try:
                this_task = this_task.__of__(self)
                status = this_task() # perform the task
            except:
                msg = ('The task %s call failed hard!' %
                       this_task.getDescription())
                LOG('Scheduler', ERROR, msg, error=sys.exc_info())
                continue

            if not status: # task may want to reschedule
                try:
                    status = this_task.next() # does it want to reschedule?
                    msg = 'next returned %s' % str(status)
                    LOG('Scheduler', BLATHER, msg)
                    if not status: # if not, that's ok.
                        continue
                except:
                    msg = ('The task %s next failed hard!' %
                           this_task.getDescription())
                    LOG('Scheduler', ERROR, msg, error=sys.exc_info())
                    continue
            try:
                next_time, next_task = status
                # make sure we don't try to store an acquisition wrapper
                next_task = getattr(next_task, 'aq_base', next_task)
            except TypeError:
                next_time, next_task = None, None
            
            if IScheduledEvent.isImplementedBy(next_task):
                self.schedule(next_time, next_task)
                continue

            msg = ('The task %s declined to reschedule itself '
                   'by returning %s from its __call__'
                   % (this_task.getDescription(), str(status)))
            LOG('Scheduler', BLATHER, msg)
                
    security.declareProtected(VIEW_SCHEDULE_PERM, 'getPendingTasks')
    def getPendingTasks(self, when=None):
        if when is None: when = time.time()
        when = int(when)
        # XXX:  We should really be storing paths and dereferencing them
        #       here;  re-wrapping the task object in the schedule is icky!
        tasks = self.tasks.items(None, when) #min, max
        return [ ( x[0], aq_base( x[1] ).__of__( self ) ) for x in tasks ]
                
    security.declareProtected(VIEW_SCHEDULE_PERM, 'getPendingTaskInfo')
    def getPendingTaskInfo(self, when=None):
        """
            Return a sequence of mappings for use by UI.
        """
        return [ ( x[0], { 'when' : x[1].getTime()
                       , 'info' : x[1].getInfo()
                       , 'description' : x[1].getDescription() } )
                 for x in self.getPendingTasks( when ) ]

    security.declareProtected(CHANGE_SCHEDULE_PERM, 'schedule')
    def schedule(self, time, task):
        LOG('Scheduler',INFO,"scheduling task = [%s] at time = [%s]" % (repr(task),str(time)))
        time = int(time)
        while not self.tasks.insert(time, task):
            # give the task as close as possible time to its requested
            # runtime
            time = time + 1

    security.declareProtected(CHANGE_SCHEDULE_PERM, 'deschedule')
    def deschedule(self, time):
        """ deschedule the task by removing from the tasks BTree """
        del self.tasks[int(time)]
            

    security.declareProtected(CHANGE_SCHEDULE_PERM, 'manage_scheduleTask')
    def manage_scheduleTask(self, description, when, path, interval,
                            max_age, max_retries, retry_backoff_time,
                            REQUEST=None):
        """ """
        task = Task(description, when, path, interval, max_age, max_retries,
                     retry_backoff_time)
        self.schedule(when, task)
        if REQUEST is not None:
            return self.manage_current_tasks(self, REQUEST)

    security.declareProtected(NOTIFY_SCHEDULE_PERM, 'manage_notifyTasks')
    def manage_notifyTasks(self, REQUEST=None):

        """ Run any pending tasks (allow poking via the ZMI)."""
        self.notify()

        if REQUEST is not None:
            return self.manage_current_tasks(self, REQUEST)

    security.declareProtected(CHANGE_SCHEDULE_PERM, 'manage_descheduleTask')
    def manage_descheduleTask(self, when, REQUEST=None):
        """ """
        if isinstance(when, IntType):
            when = [when]
        for t in when:
            try:
                # this try-except is in case the form is submitted after
                # notify has kicked a task out of the queue.
                self.deschedule(t)
            except KeyError:
                pass
        if REQUEST is not None:
            return self.manage_current_tasks(self, REQUEST)

    security.declarePublic('format_time')
    def format_time(self, t):
        return time.ctime(t)

    security.declarePublic('now')
    def now(self):
        return int(time.time())

    security.declareProtected(MGMT_SCREEN_PERM, 'manage_current_tasks')
    manage_current_tasks = PageTemplateFile(
        'www/manage_current_tasks', globals(),__name__='manage_current_tasks'
        )

    security.declareProtected(MGMT_SCREEN_PERM, 'manage_schedule_tasks')
    manage_schedule_tasks = PageTemplateFile(
       'www/manage_schedule_tasks',globals(),__name__='manage_schedule_tasks'
        )

    # FIXME put up a 'standard' dedicated member for this script
    #       instead of hard-coded specific admin
    security.declareProtected(View, 'getClockCommand')
    def getClockCommand(self,credentials="anonymous:nopass"):
        if credentials.find("'") != -1:
            raise ValueError, "authentication must *not* contain << ' >> characters "
        # IMPORTANT: Assume Z2 server is launched
        # using ./bin/python z2.py [options] 
        # and at zope root directory
        return sys.executable + " " + os.path.join(
                sys.exec_prefix, "Products","Scheduler","clock.py"
            ) + (" -U %s -a '%s'"  % (
                self.absolute_url(),
                credentials, 
            ) )

    security.declareProtected(View, 'startScheduler')
    def startScheduler(self,force=0):
        " Starts external scheduling process "
        mycredentials = str(self.login)+":"+str(self.password)
        command = self.getClockCommand(mycredentials)
        if not self.isSchedulerStarted():
            self._v_scheduler_clock = Popen3(command)
        debug("Scheduler started at pid = [%s] using command = [%s]"%(self._v_scheduler_clock.pid, command))

    security.declareProtected(View, 'isSchedulerStarted')
    def isSchedulerStarted(self):
        " Returns: true if scheduler is started "
        return (hasattr(self,"_v_scheduler_clock") and self._v_scheduler_clock)
        
    security.declareProtected(View, 'stopScheduler')
    def stopScheduler(self,force=0):
        " Stops external scheduling process "
        if force or self.isSchedulerStarted():
            pid = self._v_scheduler_clock.pid
            os.kill(pid, SIGTERM)
            debug("Scheduler stopped at pid = [%s]"%str(pid))
        self._v_scheduler_clock = None
        
    security.declareProtected(View, 'restartScheduler')
    def restartScheduler(self,force=0):
        " Restart external scheduling process "
        self.stopScheduler(force=force)
        self.startScheduler(force=force)
        
    security.declarePublic('schedule_slaves')
    def schedule_slaves(self):
        " Dispatcher method "
        debug("***** received TOP from clock *****")

    security.declareProtected(View, 'manage_clock')
    manage_clock = PageTemplateFile(
            'skins/clock_edit.pt', globals(), __name__='manage_clock'
        )

class Task(Implicit, Persistent):

    security = ClassSecurityInfo()
    security.declareObjectPublic()
    security.setDefaultAccess("allow")

    __implements__ = IScheduledEvent

    def __init__(self, description, when, path, interval=None, max_age=600,
                 max_retries=5, retry_backoff_time=60, filter_data=None):
        self.description = description
        self.when = when
        self.path = path
        self.interval = interval
        self.max_age = max_age
        self.max_retries = max_retries
        self.retry_backoff_time = retry_backoff_time
        self.retries = 0
        self.filter_data = filter_data

    def __repr__(self):
        return "<Task description='%s' when='%s' path='%s' interval=%s ... >" % (self.description, self.when, self.path, self.interval)

    def __call__(self):
        now = int(time.time())
        # secondary max_age guard (in case of clock failure)
        if now > self.when + self.max_age:
            msg = ('Task "%s" scheduled for %s failed due to maximum '
                   'age exceeded' % (self.getDescription(),
                                     pretty_time(self.when)))
            LOG('Scheduler', ERROR, msg)
            return 1 # dont schedule a retry, we've failed
        try:
            REQUEST = getattr(self, 'REQUEST')
            result = self._runMethod(REQUEST)
        except:
            # retry guard
            if self.retries + 1 > self.max_retries:
                # dont schedule a retry, we've failed
                msg = ('Task "%s" scheduled for %s failed due to maximum '
                       'retries exceeded' % (self.description,
                                                   pretty_time(self.when)))
                LOG('Scheduler', ERROR, msg)
                return 1
            
            then = now + self.retry_backoff_time

            # primary max_age guard
            if then > self.when + self.max_age:
                # dont schedule a retry, we've failed
                msg = ('Task "%s" scheduled for %s failed due to maximum '
                       'age exceeded' % (self.description,
                                               pretty_time(self.when)))
                LOG('Scheduler', ERROR, msg)
                return 1

            # we schedule a retry if we get to here
            msg = ('Task "%s" scheduled for %s could not be completed due '
                   'to an exception, retrying at %s' % (self.description,
                                                        pretty_time(self.when),
                                                        pretty_time(then)))
            LOG('Scheduler', PROBLEM, msg, error=sys.exc_info())
            self.retries = self.retries + 1
            return then, self

        # If we get here, we log success and return None
        msg = ('Task %s was run successfully with result %s' %
               (self.description, result))
        LOG('Scheduler', BLATHER, msg)
        return None

    def next(self):
        # having an interval means the task is recurring
        if self.interval:
            now = int(time.time())
            next_time = self.when + self.interval
            skipped = 0
            while next_time < now:
                # catch up!
                skipped  = skipped + 1
                next_time = next_time + self.interval
            if skipped:
                # report that we've lost some task runs
                msg = ('Task "%s" was skipped %s times due to a clock '
                       'failure or task server downtime'  % (self.description,
                                                              skipped))
                LOG('Scheduler', BLATHER, msg)
            return next_time, self.clone(next_time)
        return None # we do not have an interval

    def info(self):
        return ""

    #sm@nuxeo.com - ven jan 17 11:57:15 CET 2003
    def getInfo(self):
        return self.info()

    def getClockPid(self):
        """ Return: the clock pid """
        if hasattr(self, "_v_scheduler_clock"):
            return self._v_scheduler_clock.pid
        return -1

    def getDescription(self):
        return self.description

    def getInfo(self):
        return ''

    def getTime(self):
        return self.when

    def getFilterData(self):
        return self.filter_data

    def _runMethod(self, request):
        args = getattr(request, 'args', {})
        method = self.restrictedTraverse(self.path)
        result = mapply(method, args, request,
                        call_object, 1,
                        missing_name,
                        dont_publish_class,
                        request, bind=1)
        return result

    def clone(self, when):
        LOG('Task', BLATHER, 'cloned task = [%s]'%str(self))
        return self.__class__(self.description,
                              when,
                              self.path,
                              self.interval,
                              self.max_age,
                              self.max_retries,
                              self.retry_backoff_time,
                              self.filter_data)

class Filter(Base, Persistent):
    def __init__(self, key):
        self.key = key
        
    def __call__(self, event):
        key = getattr(event, 'getFilterData', None)
        if key and key() == self.key:
            return 1
        return 0

def pretty_time(t):
    return time.ctime(t)

def debug(message,summary='',severity=0):
    LOG('Scheduler',severity,summary,str(message)+'\n\n')


# Management forms

manage_addSchedulerForm = PageTemplateFile(
        'www/addScheduler',
        globals(),
        __name__='manage_addSchedulerForm'
        )

def manage_addScheduler(self, id, title, filter_data, REQUEST=None):
    """ """
    ob = Scheduler(id, title, filter_data)
    event_service = getattr(self, ob.event_service)
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)

Globals.InitializeClass(Scheduler)
Globals.InitializeClass(Task)
